import abc
import collections
import contextlib
import dataclasses
from enum import Enum
import re
import sys
import types
import typing
from typing import Any, List, Optional, Union, get_args, get_origin

__version__ = "1.2.0"

OR_OPERATOR_SUPPORTED = sys.version_info >= (3, 10)
"""Flag for whether PEP 604's | operator (bitwise or) between types is supported."""

LITERAL_TYPE_SUPPORTED = sys.version_info >= (3, 8)
"""(DEPRECATED) Flag for whether PEP 586's typing.Literal is supported. typenames no longer 
supports Python versions where this is false.
"""

# TypeVar for type annotations
# Most typing special forms have type 'object'
# There is a stalled proposal for TypeForm: https://github.com/python/mypy/issues/9773
if sys.version_info >= (3, 10):
    _TypeForm = typing.TypeVar("_TypeForm", bound=type | types.UnionType | object)
else:
    _TypeForm = typing.TypeVar("_TypeForm", bound=typing.Union[type, object])


class UnionSyntax(str, Enum):
    """Enum for union syntax options. See "Union Syntax" section of README for documentation."""

    AS_GIVEN = "as_given"
    OR_OPERATOR = "or_operator"
    SPECIAL_FORM = "special_form"


class OptionalSyntax(str, Enum):
    """Enum for optional syntax options. See "Optional Syntax" section of README for
    documentation."""

    AS_GIVEN = "as_given"
    OR_OPERATOR = "or_operator"
    OPTIONAL_SPECIAL_FORM = "optional_special_form"
    UNION_SPECIAL_FORM = "union_special_form"


class StandardCollectionSyntax(str, Enum):
    """Enum for standard collection parameterized generic options. See "Standard Collection Syntax"
    section of README for documentation."""

    AS_GIVEN = "as_given"
    STANDARD_CLASS = "standard_class"
    TYPING_MODULE = "typing_module"


DEFAULT_REMOVE_MODULES: List[Union[str, re.Pattern]] = [
    "__main__",
    "builtins",
    re.compile(r"^collections\.(abc\.)?"),
    "contextlib",
    "re",
    "types",
    "typing",
]
"""List of standard library modules used as the default value for the remove_modules option."""

REMOVE_ALL_MODULES = [re.compile(r"^(<?\w+>?\.)+")]


@dataclasses.dataclass
class TypenamesConfig:
    """Dataclass that holds all configuration options. See "Configurable options" section of README
    for documentation."""

    union_syntax: UnionSyntax = UnionSyntax.AS_GIVEN
    optional_syntax: OptionalSyntax = OptionalSyntax.AS_GIVEN
    standard_collection_syntax: StandardCollectionSyntax = StandardCollectionSyntax.AS_GIVEN
    remove_modules: List[Union[str, re.Pattern]] = dataclasses.field(
        default_factory=lambda: list(DEFAULT_REMOVE_MODULES)
    )

    def __post_init__(self):
        self.union_syntax = UnionSyntax(self.union_syntax)
        self.optional_syntax = OptionalSyntax(self.optional_syntax)
        self.standard_collection_syntax = StandardCollectionSyntax(self.standard_collection_syntax)

    @property
    def remove_modules_patterns(self) -> typing.Iterator[re.Pattern]:
        """Generator that yields remove_modules configuration values as compiled regex
        patterns."""
        for module in self.remove_modules:
            if isinstance(module, re.Pattern):
                yield module
            else:
                yield re.compile(r"^{}\.".format(module.replace(".", r"\.")))


@dataclasses.dataclass(repr=False)
class BaseNode(abc.ABC, typing.Generic[_TypeForm]):
    """Abstract base class for a typenames node."""

    tp: _TypeForm
    config: TypenamesConfig

    @property
    @abc.abstractmethod
    def is_none_type(self) -> bool:
        """Whether this node corresponds to NoneType, which is the type of None. Used to identify
        the "Optional" special case of a union of types.

        Returns:
            bool: True if this node's tp is NoneType.
        """

    def process_module_prefix(self, module_prefix: str) -> str:
        """Processes a module prefix (including the trailing '.') according to the 'remove_modules'
        settings in the given configuration."""
        # Remove module names
        for pattern in self.config.remove_modules_patterns:
            module_prefix = pattern.sub("", module_prefix)
        return module_prefix

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {repr(self.tp)}>"


@dataclasses.dataclass(repr=False)
class TypeNode(BaseNode):
    """Node that represents a singleton type."""

    @property
    def is_none_type(self) -> bool:
        return self.tp is type(None)

    def __str__(self) -> str:
        if hasattr(self.tp, "__module__"):
            module_prefix = getattr(self.tp, "__module__") + "."
        else:
            module_prefix = ""

        if self.tp is Ellipsis:
            type_name = "..."
        elif self.tp is type(None):
            type_name = "None"
        elif self.tp is typing.Any:
            type_name = "Any"
        elif isinstance(self.tp, typing.ForwardRef):
            forward_arg = self.tp.__forward_arg__
            # Assume if there is dotted path, then it is a module path
            if "." in forward_arg:
                module, dot, type_name = forward_arg.rpartition(".")
                module_prefix = module + dot
            else:
                type_name = forward_arg
        else:
            type_name = getattr(self.tp, "__qualname__", repr(self.tp))
        for pattern in self.config.remove_modules_patterns:
            type_name = pattern.sub("", type_name)
        # Remove module names
        module_prefix = self.process_module_prefix(module_prefix)

        return module_prefix + type_name


@dataclasses.dataclass(repr=False)
class GenericNode(BaseNode):
    """Node that represents a generic type."""

    origin: Union[Any, None]
    arg_nodes: List[BaseNode]

    @property
    def is_none_type(self) -> bool:
        return False

    def __str__(self) -> str:
        arg_nodes = list(self.arg_nodes)

        # Case: typing.Union
        if is_union_special_form(self.tp):
            # Case: typing.Optional
            if any(a.is_none_type for a in arg_nodes):
                if self.config.optional_syntax == OptionalSyntax.OR_OPERATOR:
                    return " | ".join(str(a) for a in arg_nodes)
                elif self.config.optional_syntax == OptionalSyntax.UNION_SPECIAL_FORM:
                    origin_module_prefix = "typing."
                    origin_name = "Union"
                else:
                    origin_module_prefix = "typing."
                    origin_name = "Optional"
                    arg_nodes = [a for a in arg_nodes if a.tp is not type(None)]
                    if len(arg_nodes) > 1:
                        # typing.Optional is only valid for a single parameter,
                        # need to use a union inside
                        arg_nodes = [
                            GenericNode(
                                tp=typing.Union[tuple(a.tp for a in arg_nodes)],  # type: ignore[arg-type]
                                config=self.config,
                                origin=typing.Union,
                                arg_nodes=arg_nodes,
                            )
                        ]

            # Case: regular Union
            else:
                if self.config.union_syntax == UnionSyntax.OR_OPERATOR:
                    return " | ".join(str(a) for a in arg_nodes)
                else:
                    origin_module_prefix = "typing."
                    origin_name = "Union"
        # Case: Union with | operator (bitwise or)
        elif is_union_or_operator(self.tp):
            is_optional = any(a.is_none_type for a in arg_nodes)
            # Case: ... | None (optional) and configured to use typing.Optional
            if is_optional and self.config.optional_syntax == OptionalSyntax.OPTIONAL_SPECIAL_FORM:
                origin_module_prefix = "typing."
                origin_name = "Optional"
                arg_nodes = [a for a in arg_nodes if a.tp is not type(None)]
                if len(arg_nodes) > 1:
                    # typing.Optional is only valid for a single parameter,
                    # need to use a union inside
                    arg_nodes = [
                        GenericNode(
                            tp=typing.Union[tuple(a.tp for a in arg_nodes)],  # type: ignore[arg-type]
                            config=self.config,
                            origin=typing.Union,
                            arg_nodes=arg_nodes,
                        )
                    ]
            # Case: ... | None (optional) and configured to use typing.Union
            elif is_optional and self.config.optional_syntax == OptionalSyntax.UNION_SPECIAL_FORM:
                origin_module_prefix = "typing."
                origin_name = "Union"
            # Case: regular union
            else:
                if self.config.union_syntax == UnionSyntax.SPECIAL_FORM:
                    origin_module_prefix = "typing."
                    origin_name = "Union"
                else:
                    return " | ".join(str(a) for a in arg_nodes)
        # Case: Standard collection class alias
        elif is_standard_collection_type_alias(self.tp):
            if self.config.standard_collection_syntax == StandardCollectionSyntax.TYPING_MODULE:
                typing_alias = STANDARD_COLLECTION_TO_TYPING_ALIAS_MAPPING[get_origin(self.tp)]
                origin_module_prefix = "typing."
                origin_name = f"{typing_alias._name}"  # type: ignore
            else:
                origin_module_prefix = self.origin.__module__ + "."
                origin_name = self.origin.__qualname__  # type: ignore[union-attr]
        # Case: Typing module collection alias
        elif is_typing_module_collection_alias(self.tp):
            if self.config.standard_collection_syntax == StandardCollectionSyntax.STANDARD_CLASS:
                origin_module_prefix = self.origin.__module__ + "."
                origin_name = self.origin.__qualname__  # type: ignore[union-attr]
            else:
                origin_module_prefix = "typing."
                origin_name = self.tp._name
        # Case: Some other generic type
        else:
            if hasattr(self.origin, "__module__"):
                origin_module_prefix = self.origin.__module__ + "."
            else:
                origin_module_prefix = ""
            origin_name = getattr(
                self.origin, "__name__", getattr(self.origin, "_name", str(self.origin))
            )

        # Remove module names
        origin_module_prefix = self.process_module_prefix(origin_module_prefix)

        args_string = ", ".join(str(a) for a in arg_nodes)
        return f"{origin_module_prefix}{origin_name}[{args_string}]"

    def __repr__(self) -> str:
        args_repr = ", ".join(repr(a) for a in self.arg_nodes)
        return f"<{type(self).__name__} {repr(self.origin)}[{args_repr}]>"


@dataclasses.dataclass(repr=False)
class ParamsListNode(BaseNode):
    """Node that represents a list of parameters. Used for the arguments of Callable."""

    arg_nodes: List[BaseNode]

    @property
    def is_none_type(self) -> typing.NoReturn:
        # Don't expect to call this—only used for checking args of union
        raise NotImplementedError(  # pragma: no cover
            "ParamsListNode.is_none_type should not be called."
        )

    def __str__(self) -> str:
        args_string = ", ".join(str(a) for a in self.arg_nodes)
        return f"[{args_string}]"


@dataclasses.dataclass(repr=False)
class LiteralNode(BaseNode):
    """Node that represents a literal value. Used for the arguments of Literal. Valid types for
    a literal value include ints, byte strings, unicode strings, bools, Enum values, None.
    """

    @property
    def is_none_type(self) -> typing.NoReturn:
        # Don't expect to call this—only used for checking args of union
        raise NotImplementedError(  # pragma: no cover
            "LiteralNode.is_none_type should not be called."
        )

    def __str__(self) -> str:
        if isinstance(self.tp, Enum):
            return f"{self.tp.__class__.__name__}.{self.tp.name}"
        else:
            return repr(self.tp)


def parse_type_tree(
    tp: _TypeForm, config: Optional[TypenamesConfig] = None, **kwargs: Any
) -> BaseNode:
    """Parses a given type annotation into a tree data structure.

    Args:
        tp (type): Type annotation
        config (Optional[TypenamesConfig]): Configuration dataclass. Defaults to None, which will
            instantiate one with default values.
        **kwargs: Override configuration options on provided or default configuration. See
            "Configurable options" section of README for documentation.

    Returns:
        BaseNode: Root node of parsed type tree
    """
    if config is None:
        config = TypenamesConfig(**kwargs)
    else:
        config = dataclasses.replace(config, **kwargs)

    node: BaseNode
    origin = get_origin(tp)
    if origin:
        args = get_args(tp)
        node = GenericNode(
            tp=tp,
            origin=origin,
            arg_nodes=[parse_type_tree(a, config=config) for a in args],
            config=config,
        )
    elif isinstance(tp, list):
        # This is the parameter list for Callable
        node = ParamsListNode(tp=tp, arg_nodes=[parse_type_tree(a) for a in tp], config=config)
    elif isinstance(tp, (int, bytes, str, Enum, bool)) or tp is None:
        node = LiteralNode(tp=tp, config=config)
    else:
        node = TypeNode(tp=tp, config=config)
    return node


def typenames(tp: _TypeForm, config: Optional[TypenamesConfig] = None, **kwargs: Any) -> str:
    """Render a string representation of a type annotation.

    Args:
        tp (type): Type annotation.
        config (Optional[TypenamesConfig]): Configuration dataclass. Defaults to None, which will
            instantiate one with default values.
        **kwargs: Override configuration options on provided or default configuration. See
            "Configurable options" section of README for documentation.

    Returns:
        str: String representation of input type.
    """
    tree = parse_type_tree(tp=tp, config=config, **kwargs)
    return str(tree)


def is_union_special_form(tp: _TypeForm) -> bool:
    """Check if type annotation is a union and uses the typing.Union special form."""
    return get_origin(tp) is typing.Union


def is_union_or_operator(tp: _TypeForm) -> bool:
    """Check if type annotation is a union and uses | operator (bitwise or)."""
    return OR_OPERATOR_SUPPORTED and isinstance(tp, types.UnionType)


STANDARD_COLLECTION_TO_TYPING_ALIAS_MAPPING = {
    # builtins module
    dict: typing.Dict,
    list: typing.List,
    set: typing.Set,
    frozenset: typing.FrozenSet,
    tuple: typing.Tuple,
    # collections module
    collections.defaultdict: typing.DefaultDict,
    collections.OrderedDict: typing.OrderedDict,
    collections.ChainMap: typing.ChainMap,
    collections.Counter: typing.Counter,
    collections.deque: typing.Deque,
    # collections.abc module
    collections.abc.Awaitable: typing.Awaitable,
    collections.abc.Coroutine: typing.Coroutine,
    collections.abc.AsyncIterable: typing.AsyncIterable,
    collections.abc.AsyncIterator: typing.AsyncIterator,
    collections.abc.AsyncGenerator: typing.AsyncGenerator,
    collections.abc.Iterable: typing.Iterable,
    collections.abc.Iterator: typing.Iterator,
    collections.abc.Generator: typing.Generator,
    collections.abc.Reversible: typing.Reversible,
    collections.abc.Collection: typing.Collection,
    collections.abc.Container: typing.Container,
    collections.abc.Callable: typing.Callable,
    collections.abc.Set: typing.AbstractSet,
    collections.abc.MutableSet: typing.MutableSet,
    collections.abc.Mapping: typing.Mapping,
    collections.abc.MutableMapping: typing.MutableMapping,
    collections.abc.Sequence: typing.Sequence,
    collections.abc.MutableSequence: typing.MutableSequence,
    collections.abc.ByteString: typing.ByteString,
    collections.abc.MappingView: typing.MappingView,
    collections.abc.KeysView: typing.KeysView,
    collections.abc.ItemsView: typing.ItemsView,
    collections.abc.ValuesView: typing.ValuesView,
    # contextlib module
    contextlib.AbstractContextManager: typing.ContextManager,
    contextlib.AbstractAsyncContextManager: typing.AsyncContextManager,
    # re module
    re.Pattern: typing.Pattern,
    re.Match: typing.Match,
}
"""Mapping from standard collection types that support use as a generic type starting in
Python 3.9 (PEP 585) to their associated typing module generic alias."""

STANDARD_COLLECTION_CLASSES = frozenset(STANDARD_COLLECTION_TO_TYPING_ALIAS_MAPPING.keys())
"""Frozenset of standard collection classes that support use as a generic type starting in
Python 3.9 (PEP 585)."""


def is_standard_collection_type_alias(tp: type) -> bool:
    """Check if type annotation is a generic type and uses a standard collection type as a generic
    alias."""
    return (
        get_origin(tp) in STANDARD_COLLECTION_CLASSES and type(tp) is not typing._GenericAlias  # type: ignore[attr-defined]
    )


def is_typing_module_collection_alias(tp: type) -> bool:
    """Check if type annotation is a generic type and uses a typing module collection generic
    alias, e.g., typing.List."""
    return (
        get_origin(tp) in STANDARD_COLLECTION_CLASSES and type(tp) is typing._GenericAlias  # type: ignore[attr-defined]
    )



import typing

from typenames import parse_type_tree, typenames

typenames(int)
# Unions
typenames(str | int)
typenames(typing.Union[str, int])
typenames(typing.Optional[str])
# Generics
typenames(dict[str, int])
typenames(typing.Dict[str, int])
# Other special forms
typenames(typing.Any)
typenames(typing.Literal["foo"])

parse_type_tree(int)
# Unions
parse_type_tree(str | int)
parse_type_tree(typing.Union[str, int])
parse_type_tree(typing.Optional[str])
# Generics
parse_type_tree(dict[str, int])
parse_type_tree(typing.Dict[str, int])
# Other special forms
parse_type_tree(typing.Any)
parse_type_tree(typing.Literal["foo"])



import collections
import enum
import sys
import typing

import pytest

from typenames import (
    DEFAULT_REMOVE_MODULES,
    REMOVE_ALL_MODULES,
    TypenamesConfig,
    is_standard_collection_type_alias,
    is_typing_module_collection_alias,
    is_union_or_operator,
    is_union_special_form,
    parse_type_tree,
    typenames,
)

T = typing.TypeVar("T")


class MyClass:
    pass


class OuterClass:
    class InnerClass:
        pass


class MyGeneric(typing.Generic[T]):
    def __init__(self, x: T):
        self.x = x


class MyEnum(enum.Enum):
    MEMBER1 = "member1"
    MEMBER2 = "member2"


class MyTypedDict(typing.TypedDict):
    x: int
    y: int
    label: str


cases = [
    (int, "int"),
    (typing.List[int], "List[int]"),
    (typing.Tuple[str, int], "Tuple[str, int]"),
    (typing.Optional[int], "Optional[int]"),
    (MyClass, "test_typenames.MyClass"),
    (OuterClass.InnerClass, "test_typenames.OuterClass.InnerClass"),
    (typing.List[MyClass], "List[test_typenames.MyClass]"),
    (typing.Optional[typing.List[MyClass]], "Optional[List[test_typenames.MyClass]]"),
    (typing.Union[float, int], "Union[float, int]"),
    (typing.Dict[str, int], "Dict[str, int]"),
    (typing.Any, "Any"),
    (typing.Dict[str, typing.Any], "Dict[str, Any]"),
    (typing.Callable[..., str], "Callable[..., str]"),
    (typing.Callable[[int], str], "Callable[[int], str]"),
    (typing.Callable[[int, float], str], "Callable[[int, float], str]"),
    (MyGeneric[int], "test_typenames.MyGeneric[int]"),
    (MyEnum, "test_typenames.MyEnum"),
    (typing.List["int"], "List[int]"),
    (typing.List["typing.Any"], "List[Any]"),
    (typing.List["enum.Enum"], "List[enum.Enum]"),
    (typing.List["MyClass"], "List[MyClass]"),
    (typing.List["OuterClass.InnerClass"], "List[OuterClass.InnerClass]"),
    # Python 3.8 adds typing.Literal, typing.Final, typing.TypedDict
    (typing.Literal["s", 0, MyEnum.MEMBER1], "Literal['s', 0, MyEnum.MEMBER1]"),
    (typing.Final[int], "Final[int]"),
    (MyTypedDict, "test_typenames.MyTypedDict"),
]


if sys.version_info >= (3, 9):
    # Python 3.9 adds use of standard collection types as a generic in type annotations (PEP 585),
    # typing.Annotated
    cases.extend(
        [
            (list[int], "list[int]"),
            (list[tuple[int, str]], "list[tuple[int, str]]"),
            (list[typing.Tuple[int, str]], "list[Tuple[int, str]]"),
            (
                list[collections.defaultdict[str, list[int]]],
                "list[defaultdict[str, list[int]]]",
            ),
            (collections.abc.Mapping[str, str], "Mapping[str, str]"),
        ]
    )

if sys.version_info >= (3, 10):
    # Python 3.10 adds union syntax with the | operator (bitwise or),
    # typing.Concatenate, typing.ParamSpec, typing.TypeAlias
    cases.extend(
        [
            (int | str, "int | str"),
            (int | None, "int | None"),
            (None | int, "None | int"),
            (int | None | str, "int | None | str"),
        ]
    )

if sys.version_info >= (3, 11):
    # Python 3.11 adds typing.LiteralString, typing.Never, typing.Self
    cases.extend(
        [
            (typing.LiteralString, "LiteralString"),
            (typing.Never, "Never"),
            (typing.Self, "Self"),
        ]
    )


@pytest.mark.parametrize("case", cases, ids=[c[1] for c in cases])
def test_typenames(case):
    assert typenames(case[0]) == case[1]


def test_remove_modules():
    # Simulate a class from another module
    class OtherModuleClass:
        __qualname__ = "OtherModuleClass"
        __module__ = "other_module"

    # Class defined in a function scope
    class FnScopeClass: ...

    # Default removal
    assert typenames(typing.Any) == "Any"
    assert typenames(MyClass) == "test_typenames.MyClass"
    assert typenames(OtherModuleClass) == "other_module.OtherModuleClass"
    assert typenames(FnScopeClass) == "test_typenames.test_remove_modules.<locals>.FnScopeClass"

    # Override
    config = TypenamesConfig(remove_modules=["test_typenames"])
    assert typenames(typing.Any, config=config) == "typing.Any"
    assert typenames(MyClass, config=config) == "MyClass"
    assert typenames(OtherModuleClass, config=config) == "other_module.OtherModuleClass"
    assert typenames(FnScopeClass, config=config) == "test_remove_modules.<locals>.FnScopeClass"

    # Add to defaults
    config = TypenamesConfig(remove_modules=DEFAULT_REMOVE_MODULES + ["test_typenames"])
    assert typenames(typing.Any, config=config) == "Any"
    assert typenames(MyClass, config=config) == "MyClass"
    assert typenames(OtherModuleClass, config=config) == "other_module.OtherModuleClass"
    assert typenames(FnScopeClass, config=config) == "test_remove_modules.<locals>.FnScopeClass"

    # All modules
    config = TypenamesConfig(remove_modules=REMOVE_ALL_MODULES)
    assert typenames(typing.Any, config=config) == "Any"
    assert typenames(MyClass, config=config) == "MyClass"
    assert typenames(OtherModuleClass, config=config) == "OtherModuleClass"
    assert typenames(OtherModuleClass, config=config) == "OtherModuleClass"
    assert typenames(FnScopeClass, config=config) == "FnScopeClass"

    if sys.version_info >= (3, 9):
        assert typenames(collections.Counter[str], config=config) == "Counter[str]"
        assert typenames(collections.abc.Sequence[str], config=config) == "Sequence[str]"


def test_union_syntax_or_operator():
    """Test that forcing optional_syntax='or_operator' results in ... | ...."""
    assert typenames(typing.Union[int, str], union_syntax="or_operator") == "int | str"

    if sys.version_info >= (3, 10):
        assert typenames(int | str, union_syntax="or_operator") == "int | str"


def test_union_syntax_union_special_form():
    """Test that forcing optional_syntax='union_special_form' results in Union[...]"""
    assert typenames(typing.Union[int, str], union_syntax="special_form") == "Union[int, str]"

    if sys.version_info >= (3, 10):
        assert typenames(int | str, union_syntax="special_form") == "Union[int, str]"


def test_optional_syntax_or_operator():
    """Test that forcing optional_syntax='or_operator' results in ... | None."""
    assert typenames(typing.Optional[int], optional_syntax="or_operator") == "int | None"

    if sys.version_info >= (3, 10):
        assert typenames(int | None, union_syntax="or_operator") == "int | None"
        assert typenames(None | int, union_syntax="or_operator") == "None | int"


def test_optional_syntax_optional_special_form():
    """Test that forcing optional_syntax='optional_special_form' results in Optional[...]."""
    assert (
        typenames(typing.Optional[int], optional_syntax="optional_special_form") == "Optional[int]"
    )

    if sys.version_info >= (3, 10):
        assert typenames(int | None, optional_syntax="optional_special_form") == "Optional[int]"
        assert typenames(None | int, optional_syntax="optional_special_form") == "Optional[int]"


def test_optional_syntax_union_special_form():
    """Test that forcing optional_syntax='union_special_form' results in Union[...]."""
    assert (
        typenames(typing.Optional[int], optional_syntax="union_special_form") == "Union[int, None]"
    )

    if sys.version_info >= (3, 10):
        assert typenames(int | None, optional_syntax="union_special_form") == "Union[int, None]"
        assert typenames(None | int, optional_syntax="union_special_form") == "Union[None, int]"


def test_optional_multiple_params():
    """Test case that a type annotation resolves to the optional case but with multiple non-None
    parameters."""
    assert typenames(typing.Optional[typing.Union[int, str]]) == "Optional[Union[int, str]]"

    if sys.version_info >= (3, 10):
        assert (
            typenames(int | str | None, optional_syntax="optional_special_form")
            == "Optional[Union[int, str]]"
        )
        assert (
            typenames(int | None | str, optional_syntax="optional_special_form")
            == "Optional[Union[int, str]]"
        )

        # union_syntax='or_operator'
        assert (
            typenames(
                int | None | str,
                optional_syntax="optional_special_form",
                union_syntax="or_operator",
            )
            == "Optional[int | str]"
        )


def test_standard_collection_syntax_standard_class():
    """Test that forcing standard_collection_syntax='standard_class' results in standard class
    output."""
    assert typenames(typing.List[int], standard_collection_syntax="standard_class") == "list[int]"

    if sys.version_info >= (3, 9):
        assert typenames(list[int], standard_collection_syntax="standard_class") == "list[int]"


def test_standard_collection_syntax_typing_module():
    """Test that forcing standard_collection_syntax='typing_module' results in typing module
    generic aliases."""
    assert typenames(typing.List[int], standard_collection_syntax="typing_module") == "List[int]"

    if sys.version_info >= (3, 9):
        assert typenames(list[int], standard_collection_syntax="typing_module") == "List[int]"


def test_node_repr():
    assert repr(parse_type_tree(int)) == "<TypeNode <class 'int'>>"
    assert repr(parse_type_tree(typing.Any)) == "<TypeNode typing.Any>"
    assert (
        repr(parse_type_tree(typing.Optional[int]))
        == "<GenericNode typing.Union[<TypeNode <class 'int'>>, <TypeNode <class 'NoneType'>>]>"
    )
    assert (
        repr(parse_type_tree(typing.Literal["a", "b"]))
        == "<GenericNode typing.Literal[<LiteralNode 'a'>, <LiteralNode 'b'>]>"
    )


is_union_special_form_cases = [
    (typing.Union[int, str], True),
    (typing.Optional[int], True),
    (typing.Union[typing.List[int], typing.Tuple[str, float]], True),
    (typing.Union[typing.Union[int, str], float], True),
    (typing.List[int], False),
    (typing.List[typing.Union[int, str]], False),
    (str, False),
    (type("Union", (), {}), False),
]


@pytest.mark.parametrize(
    "case",
    is_union_special_form_cases,
    ids=[str(c[0]) for c in is_union_special_form_cases],
)
def test_is_union_special_form(case):
    """Test that is_union_special_form correctly identifies if using typing.Union."""
    assert is_union_special_form(case[0]) == case[1]


@pytest.mark.parametrize(
    "case",
    is_union_special_form_cases,
    ids=[str(c[0]) for c in is_union_special_form_cases],
)
def test_is_union_or_operator_for_typing_alias_cases(case):
    """Test that is_union_or_operator correctly returns False for all typing.Union test cases."""
    assert is_union_or_operator(case[0]) is False


if sys.version_info >= (3, 10):
    # Python 3.10 adds | operator (bitwise or) for Union
    is_union_or_operator_cases = [
        (int | str, True),
        (int | None, True),
        (None | int, True),
        (int | str | float, True),
        (list[int] | tuple[str, float], True),
        (list[int | str], False),
    ]

    @pytest.mark.parametrize(
        "case",
        is_union_or_operator_cases,
        ids=[str(c[0]) for c in is_union_or_operator_cases],
    )
    def test_is_union_or_operator(case):
        """Test that is_union_or_operator correctly identifies cases using | operator."""
        assert is_union_or_operator(case[0]) == case[1]

    @pytest.mark.parametrize(
        "case",
        is_union_or_operator_cases,
        ids=[str(c[0]) for c in is_union_or_operator_cases],
    )
    def test_is_union_special_form_for_or_operator_cases(case):
        """Test that is_union_special_form correctly returns False for all | operator cases."""
        assert is_union_special_form(case[0]) is False


is_typing_module_collection_alias_cases = [
    (typing.List[int], True),
    (typing.Dict[str, int], True),
    (typing.Tuple[int, int], True),
    (typing.List[typing.Optional[int]], True),
    (typing.Dict[typing.Tuple[int, int], str], True),
    (int, False),
    (str, False),
    (typing.Optional[typing.List[int]], False),
    (typing.Union[typing.List[int], typing.List[str]], False),
    (collections.namedtuple("Point", ["x", "y"]), False),
]


@pytest.mark.parametrize(
    "case",
    is_typing_module_collection_alias_cases,
    ids=[str(c[0]) for c in is_typing_module_collection_alias_cases],
)
def test_is_typing_module_collection_alias(case):
    assert is_typing_module_collection_alias(case[0]) == case[1]


@pytest.mark.parametrize(
    "case",
    is_typing_module_collection_alias_cases,
    ids=[str(c[0]) for c in is_typing_module_collection_alias_cases],
)
def test_is_standard_collection_type_alias_for_typing_alias_cases(case):
    """Test that is_standard_collection_type_alias correctly returns False for all typing module
    alias cases."""
    assert is_standard_collection_type_alias(case[0]) is False


if sys.version_info >= (3, 9):
    # Python 3.9 adds generic aliases for standard collections
    is_standard_collection_type_alias_cases = [
        (list[int], True),
        (dict[str, int], True),
        (tuple[int, int], True),
        (list[typing.Optional[int]], True),
        (dict[tuple[int, int], str], True),
        (dict[typing.Tuple[int, int], str], True),
        (typing.Optional[list[int]], False),
        (typing.Union[list[int], list[str]], False),
    ]

    @pytest.mark.parametrize(
        "case",
        is_standard_collection_type_alias_cases,
        ids=[str(c[0]) for c in is_standard_collection_type_alias_cases],
    )
    def test_is_standard_collection_type_alias(case):
        assert is_standard_collection_type_alias(case[0]) == case[1]

    @pytest.mark.parametrize(
        "case",
        is_standard_collection_type_alias_cases,
        ids=[str(c[0]) for c in is_standard_collection_type_alias_cases],
    )
    def test_is_typing_module_collection_alias_for_collection_types(case):
        """Test that is_typing_module_collection_alias correctly returns False for all standard
        collection type aliases."""
        assert is_typing_module_collection_alias(case[0]) is False

    def test_nested_collection_types_nested_both():
        """Test that is_typing_module_collection_alias and is_standard_collection_type_alias
        correctly identify nested cases using both typing approaches."""
        tp1 = typing.Dict[tuple[int, int], str]
        assert is_typing_module_collection_alias(tp1) is True
        assert is_standard_collection_type_alias(tp1) is False

        tp2 = dict[typing.Tuple[int, int], str]
        assert is_typing_module_collection_alias(tp2) is False
        assert is_standard_collection_type_alias(tp2) is True
